import {
    ArgumentsHost,
    BadGatewayException,
    BadRequestException,
    ConflictException,
    ExceptionFilter,
    ForbiddenException,
    GatewayTimeoutException,
    GoneException,
    HttpStatus,
    HttpVersionNotSupportedException,
    ImATeapotException,
    InternalServerErrorException,
    MethodNotAllowedException,
    NotAcceptableException,
    NotFoundException,
    NotImplementedException,
    PayloadTooLargeException,
    PreconditionFailedException,
    RequestTimeoutException,
    ServiceUnavailableException,
    UnauthorizedException,
    UnprocessableEntityException,
    UnsupportedMediaTypeException,
} from '@nestjs/common';
import { mongo } from 'mongoose';
import { ValidationError } from 'class-validator';
import { HandledException } from '../exceptions/custom.exception';
import { RESTError } from '../models/dtos/wrappers/rest-wrapper.class';

export class BaseExceptionsFilter implements ExceptionFilter {
    private _debug: boolean = false;
    private _log: boolean = false;

    constructor(debug: boolean, log?: boolean) {
        this._debug = debug;
        this._log = log;
    }

    catch(exception: any, host: ArgumentsHost) {
        const context = host.switchToHttp();
        const response = context.getResponse<Response>();

        if (this._debug) {
            console.log('BaseExceptionsFilter - name:', exception.constructor.name);
        }

        if (this._log) {
            console.log(exception);
        }

        let handledException: HandledException | any = null;

        const { restErrors, statusCode } = this.filterException(exception);
        handledException = new HandledException(restErrors, statusCode);

        handledException.send(response);
    }

    private filterException(exception: any) {
        const restErrors: RESTError[] = [];
        let statusCode: HttpStatus = null;

        try {
            if (exception instanceof HandledException) {
                restErrors.push(...exception.getErrors());
                statusCode = exception.getStatus() || HttpStatus.BAD_REQUEST;
            } else if (exception instanceof RESTError) {
                restErrors.push(exception);
                statusCode = exception.statusCode || HttpStatus.BAD_REQUEST;
            } else if (exception instanceof Array) {
                exception.forEach((item) => {
                    restErrors.push(...this.filterException(item).restErrors);
                });
                statusCode = HttpStatus.BAD_REQUEST;
            } else if (exception instanceof BadRequestException) {
                restErrors.push(new RESTError('bad_request'));
                statusCode = HttpStatus.BAD_REQUEST;
            } else if (exception instanceof UnauthorizedException) {
                restErrors.push(new RESTError('unauthorized'));
                statusCode = HttpStatus.UNAUTHORIZED;
            } else if (exception instanceof NotFoundException) {
                restErrors.push(new RESTError('method_not_found', exception.message));
                statusCode = HttpStatus.METHOD_NOT_ALLOWED;
            } else if (exception instanceof ForbiddenException) {
                restErrors.push(new RESTError('forbidden'));
                statusCode = HttpStatus.FORBIDDEN;
            } else if (exception instanceof NotAcceptableException) {
                restErrors.push(new RESTError('NotAcceptableException'));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof RequestTimeoutException) {
                restErrors.push(new RESTError('RequestTimeoutException'));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof ConflictException) {
                restErrors.push(new RESTError('ConflictException'));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof GoneException) {
                restErrors.push(new RESTError('GoneException'));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof HttpVersionNotSupportedException) {
                restErrors.push(new RESTError('HttpVersionNotSupportedException'));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof PayloadTooLargeException) {
                restErrors.push(new RESTError('PayloadTooLargeException'));
                statusCode = HttpStatus.PAYLOAD_TOO_LARGE;
            } else if (exception instanceof UnsupportedMediaTypeException) {
                restErrors.push(new RESTError('UnsupportedMediaTypeException'));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof UnprocessableEntityException) {
                restErrors.push(new RESTError('UnprocessableEntityException'));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof InternalServerErrorException) {
                restErrors.push(new RESTError('InternalServerErrorException'));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof NotImplementedException) {
                restErrors.push(new RESTError('NotImplementedException'));
                statusCode = HttpStatus.NOT_IMPLEMENTED;
            } else if (exception instanceof ImATeapotException) {
                restErrors.push(new RESTError('ImATeapotException'));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof MethodNotAllowedException) {
                restErrors.push(new RESTError('MethodNotAllowedException'));
                statusCode = HttpStatus.METHOD_NOT_ALLOWED;
            } else if (exception instanceof BadGatewayException) {
                restErrors.push(new RESTError('BadGatewayException'));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof ServiceUnavailableException) {
                restErrors.push(new RESTError('ServiceUnavailableException'));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof GatewayTimeoutException) {
                restErrors.push(new RESTError('GatewayTimeoutException'));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof PreconditionFailedException) {
                restErrors.push(new RESTError('PreconditionFailedException'));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof TypeError) {
                restErrors.push(new RESTError('type_error', exception.message));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof mongo.MongoError) {
                restErrors.push(new RESTError('mongo_error', exception.message));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof ValidationError) {
                restErrors.push(new RESTError('validation_error', exception));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else if (exception instanceof mongo.MongoServerError) {
                restErrors.push(new RESTError('mongo_server_error', exception.message));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            } else {
                restErrors.push(new RESTError('unknown_error', exception.message));
                statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
            }
        } catch (ex) {
            restErrors.push(new RESTError('base_exception_filter_error', ex));
            statusCode = HttpStatus.INTERNAL_SERVER_ERROR;
        }

        return { restErrors, statusCode };
    }
}
